from enum import auto
from typing import Callable, Tuple

import torch

try:
    # be ready for 3.10 when it drops
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum

TransformType = Callable[[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]


class EncodingType(StrEnum):
    ABSOLUTE = auto()
    RELATIVE = auto()



import logging

from nn_core.console_logging import NNRichHandler

# Required workaround because PyTorch Lightning configures the logging on import,
# thus the logging configuration defined in the __init__.py must be called before
# the lightning import otherwise it has no effect.
# See https://github.com/PyTorchLightning/pytorch-lightning/issues/1503
lightning_logger = logging.getLogger("pytorch_lightning")
# Remove all handlers associated with the lightning logger.
for handler in lightning_logger.handlers[:]:
    lightning_logger.removeHandler(handler)
lightning_logger.propagate = True

FORMAT = "%(message)s"
logging.basicConfig(
    format=FORMAT,
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        NNRichHandler(
            rich_tracebacks=True,
            show_level=True,
            show_path=True,
            show_time=True,
            omit_repeated_times=True,
        )
    ],
)

try:
    from ._version import __version__ as __version__
except ImportError:
    import sys

    print(
        "Project not installed in the current env, activate the correct env or install it with:\n\tpip install -e .",
        file=sys.stderr,
    )
    __version__ = "unknown"



import functools
from typing import Optional, Sequence

import torch
import torch.nn.functional as F
from torch import nn

from latentis.types import TransformType


class RelativeProjection(nn.Module):
    def __init__(self, name: str, func: TransformType) -> None:
        super().__init__()
        self._name: str = name
        self.func: TransformType = func

    @property
    def name(self) -> str:
        return self._name

    def forward(self, x: torch.Tensor, anchors: torch.Tensor) -> torch.Tensor:
        return self.func(x=x, anchors=anchors)


def change_of_basis(x: torch.Tensor, anchors: torch.Tensor) -> torch.Tensor:
    return torch.linalg.lstsq(anchors.T, x.T)[0].T


def cosine_dist(
    x: torch.Tensor,
    anchors: torch.Tensor,
) -> torch.Tensor:
    x = F.normalize(x, p=2, dim=-1)
    anchors = F.normalize(anchors, p=2, dim=-1)

    return x @ anchors.T


def geodesic_dist(
    x: torch.Tensor,
    anchors: torch.Tensor,
) -> torch.Tensor:
    x = F.normalize(x, p=2, dim=-1)
    anchors = F.normalize(anchors, p=2, dim=-1)

    return 1 - torch.arccos(x @ anchors.T)


def lp_dist(
    x: torch.Tensor,
    anchors: torch.Tensor,
    p: int,
) -> torch.Tensor:
    return torch.cdist(x, anchors, p=p)


def euclidean_dist(
    x: torch.Tensor,
    anchors: torch.Tensor,
) -> torch.Tensor:
    return lp_dist(x=x, anchors=anchors, p=2)


class Projections:
    COSINE = RelativeProjection(name="cosine", func=cosine_dist)
    GEODESIC = RelativeProjection(name="geodesic", func=geodesic_dist)
    COB = RelativeProjection(name="change_of_basis", func=change_of_basis)
    EUCLIDEAN = RelativeProjection(name="euclidean", func=euclidean_dist)
    L1 = RelativeProjection(name="l1", func=functools.partial(lp_dist, p=1))


class RelativeProjector(nn.Module):
    # TODO: Add support for anchor optimization
    def __init__(
        self,
        projection: RelativeProjection,
        anchors: Optional[torch.Tensor] = None,
        abs_transforms: Optional[Sequence[TransformType]] = None,
        rel_transforms: Optional[Sequence[TransformType]] = None,
    ) -> None:
        super().__init__()
        self.projection: TransformType = projection

        self.abs_transforms = nn.ModuleList(
            abs_transforms
            if isinstance(abs_transforms, Sequence)
            else []
            if abs_transforms is None
            else [abs_transforms]
        )
        self.rel_transforms = nn.ModuleList(
            rel_transforms
            if isinstance(rel_transforms, Sequence)
            else []
            if rel_transforms is None
            else [rel_transforms]
        )

        if anchors is not None:
            self.register_buffer("anchors", anchors)

    def set_anchors(self, anchors: torch.Tensor) -> "RelativeProjector":
        self.register_buffer("anchors", anchors)
        return self

    @property
    def name(self) -> str:
        return self._name

    def forward(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None) -> torch.Tensor:
        assert anchors is not None or self.anchors is not None, "anchors must be provided"
        anchors = anchors if anchors is not None else self.anchors

        # absolute normalization/transformation
        transformed_x = x
        transformed_anchors = anchors
        for abs_transform in self.abs_transforms:
            transformed_x = abs_transform(x=transformed_x, anchors=anchors)
            transformed_anchors = abs_transform(x=transformed_anchors, anchors=anchors)

        # relative projection of x with respect to the anchors
        rel_x = self.projection(x=transformed_x, anchors=transformed_anchors)

        # relative normalization/transformation
        for rel_transform in self.rel_transforms:
            # TODO: handle the case where the rel_transform needs additional arguments (e.g. rel_anchors, x, ...)
            # maybe with a custom Compose object with kwargs
            rel_x = rel_transform(x=rel_x, anchors=anchors)

        return rel_x


class PointWiseProjector(RelativeProjector):
    def __init__(
        self,
        name: str,
        func: TransformType,
        anchors: Optional[torch.Tensor] = None,
        abs_transforms: Optional[Sequence[TransformType]] = None,
        rel_transforms: Optional[Sequence[TransformType]] = None,
    ) -> None:
        super().__init__(
            name=name, projection=func, anchors=anchors, abs_transforms=abs_transforms, rel_transforms=rel_transforms
        )

    def forward(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None) -> torch.Tensor:
        assert anchors is not None or self.anchors is not None, "anchors must be provided"
        anchors = anchors if anchors is not None else self.anchors

        # absolute normalization/transformation
        transformed_x = x
        transformed_anchors = anchors
        for abs_transform in self.abs_transforms:
            transformed_x = abs_transform(x=transformed_x, anchors=anchors)
            transformed_anchors = abs_transform(x=transformed_anchors, anchors=anchors)

        # relative projection of x with respect to the anchors
        rel_x = []
        for point in x:
            partial_rel_data = []
            for anchor in transformed_anchors:
                partial_rel_data.append(self.projection(x=point, anchors=anchor))
            rel_x.append(partial_rel_data)
        rel_x = torch.as_tensor(rel_x, dtype=anchors.dtype, device=anchors.device)

        # relative normalization/transformation
        for rel_transform in self.rel_transforms:
            # TODO: handle the case where the rel_transform needs additional arguments (e.g. rel_anchors, x, ...)
            # maybe with a custom Compose object with kwargs
            rel_x = rel_transform(x=rel_x, anchors=anchors)

        return rel_x



from abc import abstractmethod
from typing import Any, Mapping, Optional, Sequence

import torch
import torch.nn.functional as F
from pytorch_lightning import seed_everything
from torch import nn

from latentis.transforms import Transform
from latentis.types import TransformType


class Translator(nn.Module):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name: str = name

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        raise NotImplementedError


class IdentityTranslator(Translator):
    def __init__(self) -> None:
        super().__init__("identity")

    def fit(self, *args, **kwargs) -> Mapping[str, Any]:
        return {}

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


class AffineTranslator(Translator):
    def __init__(self) -> None:
        super().__init__("affine")

    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        with torch.enable_grad():
            translation = nn.Linear(source_data.size(1), target_data.size(1), device=source_data.device)
            optimizer = torch.optim.Adam(translation.parameters(), lr=1e-3)

            for _ in range(100):
                optimizer.zero_grad()
                loss = F.mse_loss(translation(source_data), target_data)
                loss.backward()
                optimizer.step()
            self.translation = translation.cpu()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.translation(x)


class LSTSQTranslator(Translator):
    def __init__(self) -> None:
        super().__init__("lstsq")
        self.translation_matrix = None

    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        translation_matrix = torch.linalg.lstsq(source_data, target_data).solution
        self.translation_matrix = torch.as_tensor(translation_matrix, dtype=torch.float32, device=source_data.device)

        return {}

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x @ self.translation_matrix


class LSTSQOrthoTranslator(Translator):
    def __init__(self) -> None:
        super().__init__("lstsq_ortho")

    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        translation_matrix = torch.linalg.lstsq(source_data, target_data).solution
        U, _, Vt = torch.svd(translation_matrix)
        self.translation_matrix = U @ Vt.T

        return {}


class SVDTranslator(Translator):
    def __init__(self) -> None:
        super().__init__("svd")
        self.translation_matrix = None

    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        # padding if necessary
        if source_data.size(1) < target_data.size(1):
            padded = torch.zeros_like(target_data)
            padded[:, : source_data.size(1)] = source_data
            source_data = padded
        elif source_data.size(1) > target_data.size(1):
            padded = torch.zeros_like(source_data)
            padded[:, : target_data.size(1)] = target_data
            target_data = padded

        #  Compute the translation vector that aligns A to B using SVD.
        assert source_data.size(1) == target_data.size(1)
        u, sigma, vt = torch.svd((target_data.T @ source_data).T)
        translation_matrix = u @ vt.T

        self.translation_matrix = torch.as_tensor(translation_matrix, dtype=torch.float32, device=source_data.device)
        sigma_rank = (~sigma.isclose(torch.zeros_like(sigma), atol=1e-1).bool()).sum().item()

        return {"sigma_rank": sigma_rank}

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x @ self.translation_matrix


class LatentTranslation(nn.Module):
    def __init__(
        self,
        seed: int,
        translator: Translator,
        source_transforms: Optional[Sequence[TransformType]] = None,
        target_transforms: Optional[Sequence[TransformType]] = None,
    ) -> None:
        super().__init__()
        self.seed: int = seed
        self.translator: Translator = translator

        self.source_transforms: Sequence[Transform] = nn.ModuleList(
            source_transforms
            if isinstance(source_transforms, Sequence)
            else []
            if source_transforms is None
            else [source_transforms]
        )
        self.target_transforms: Sequence[Transform] = nn.ModuleList(
            target_transforms
            if isinstance(target_transforms, Sequence)
            else []
            if target_transforms is None
            else [target_transforms]
        )

    def fit(self, source_data: torch.Tensor, target_data: torch.Tensor) -> Mapping[str, Any]:
        seed_everything(self.seed)
        translator_info = self.translator.fit(source_data=source_data, target_data=target_data)

        # for transform in self.source_transforms:
        #     transform.fit(source_data=source_data)
        # for transform in self.target_transforms:
        #     transform.fit(target_data=target_data)
        self.register_buffer("source_data", source_data)
        self.register_buffer("target_data", target_data)

        return translator_info

    def forward(self, x: torch.Tensor, compute_info: bool = True) -> torch.Tensor:
        source_x = x
        for transform in self.source_transforms:
            source_x = transform(x=source_x, anchors=self.source_data)

        target_x = self.translator(source_x)

        for transform in reversed(self.target_transforms):
            target_x = transform.reverse(x=target_x, anchors=self.target_data)

        return {"source": source_x, "target": target_x, "info": {}}



from abc import abstractmethod
from typing import Mapping, Optional

import torch
import torch.nn.functional as F
from torch import nn


# https://github.com/scikit-learn/scikit-learn/blob/7f9bad99d6e0a3e8ddf92a7e5561245224dab102/sklearn/preprocessing/_data.py#L87
def _handle_zeros(scale: torch.Tensor, copy=True, constant_mask=None):
    """Set scales of near constant features to 1.

    The goal is to avoid division by very small or zero values.

    Near constant features are detected automatically by identifying
    scales close to machine precision unless they are precomputed by
    the caller and passed with the `constant_mask` kwarg.

    Typically for standard scaling, the scales are the standard
    deviation while near constant features are better detected on the
    computed variances which are closer to machine precision by
    construction.
    """
    eps = torch.finfo(scale.dtype).eps

    # if we are fitting on 1D tensors, scale might be a scalar
    if scale.ndim == 0:
        return 1 if scale == 0 else scale
    elif isinstance(scale, torch.Tensor):
        if constant_mask is None:
            # Detect near constant values to avoid dividing by a very small
            # value that could lead to surprising results and numerical
            # stability issues.
            constant_mask = scale < 10 * eps

        if copy:
            # New tensor to avoid side-effects
            scale = scale.clone()
        scale[constant_mask] = 1.0
        scale[scale == 0.0] = 1.0
        return scale


class Transform(nn.Module):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name: str = name
        self.fitted: bool = False

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def _fit(self, x: torch.Tensor, *args, **kwargs) -> Mapping[str, torch.Tensor]:
        raise NotImplementedError

    def fit(self, x: torch.Tensor, *args, **kwargs) -> None:
        raise NotImplementedError
        # TODO: Implement fit method
        for key, value in self._fit(x=x, *args, **kwargs).items():
            self.register_buffer(key, value)
        self.fitted: bool = True

    def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError

    def reverse(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"


class Transforms:
    class Centering(Transform):
        def __init__(self) -> None:
            super().__init__(name="centering")

        def _fit(self, anchors: torch.Tensor, *args, **kwargs) -> Mapping[str, torch.Tensor]:
            return {"shift": anchors.mean(dim=0)}

        def forward(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            shift = anchors.mean(dim=0) if anchors is not None else self.shift
            return x - shift

        def reverse(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            shift = anchors.mean(dim=0) if anchors is not None else self.shift
            return x + shift

    class STDScaling(Transform):
        def __init__(self) -> None:
            super().__init__(name="std_scaling")

        def _fit(self, anchors: torch.Tensor, *args, **kwargs) -> Mapping[str, torch.Tensor]:
            return {"scale": _handle_zeros(anchors.std(dim=0))}

        def forward(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            scale = _handle_zeros(anchors.std(dim=0)) if anchors is not None else self.scale
            return x / scale

        def reverse(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            scale = _handle_zeros(anchors.std(dim=0)) if anchors is not None else self.scale
            return x * scale

    class StandardScaling(Transform):
        def __init__(self) -> None:
            super().__init__(name="standard_scaling")

        def _fit(self, anchors: torch.Tensor, *args, **kwargs) -> Mapping[str, torch.Tensor]:
            return {"shift": anchors.mean(dim=0), "scale": _handle_zeros(anchors.std(dim=0))}

        def forward(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            shift = anchors.mean(dim=0) if anchors is not None else self.shift
            scale = _handle_zeros(anchors.std(dim=0)) if anchors is not None else self.scale

            return (x - shift) / scale

        def reverse(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            shift = anchors.mean(dim=0) if anchors is not None else self.shift
            scale = _handle_zeros(anchors.std(dim=0)) if anchors is not None else self.scale

            return (x * scale) + shift

    class L2(Transform):
        def __init__(self) -> None:
            super().__init__(name="l2")

        def _fit(self, anchors: torch.Tensor, *args, **kwargs) -> Mapping[str, torch.Tensor]:
            return {"mean_norm": anchors.norm(dim=1).mean()}

        def forward(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
            return F.normalize(x, p=2, dim=-1)

        def reverse(self, x: torch.Tensor, anchors: Optional[torch.Tensor] = None, *args, **kwargs) -> torch.Tensor:
            assert (
                anchors is not None or self.fitted
            ), "Either anchors must be provided or the transform must be fit first."

            mean_norm = anchors.norm(dim=1).mean() if anchors is not None else self.mean_norm

            return x * mean_norm
